"""
Configuration
-------------

Morelia looks for configuration file "pyproject.toml" in current directory.
If it is not present then it looks for "~/.config/morelia/config.toml".

You can bypass that setting environment variable MORELIA_CONFIG to the path to config file.

Morelia reads it's configuration from "tool.morelia.<section>" namespace.
You can give section part when calling "verify" method.

E.g. if you have configuration:

.. code-block:: toml

    [tool.morelia.default]
    wip=false

    [tool.morelia.myconfig]
    wip=true

and call "verify" as:

.. code-block:: python

   verify(filename, self, config="myconfig")

Then it would be run with wip (work in progress) mode active.

If no config is passed, then "default" is assumed.
"""

import os
from pathlib import Path

import toml

from morelia.formatters import (
    Buffered,
    FileOutput,
    RemoteOutput,
    TerminalOutput,
    TextFormat,
)
from morelia.matchers import MethodNameStepMatcher, ParseStepMatcher, RegexpStepMatcher

MATCHERS = {
    "parse": ParseStepMatcher,
    "regex": RegexpStepMatcher,
    "method": MethodNameStepMatcher,
}


FORMATTERS = {"text": TextFormat}


OUTPUTS = {"terminal": TerminalOutput, "file": FileOutput, "remote": RemoteOutput}


class TOMLConfig:
    def __init__(self, section="default", filename=None):
        self.__section = section
        path = self.__get_path(filename)
        self.__data = {"wip": False, "matchers": ["regex", "parse", "method"]}
        if path:
            data = toml.load(path).get("tool", {}).get("morelia", {}).get(section, {})
        else:
            data = {}
        self.__data.update(data)

    def __get_path(self, filename):
        if filename and Path(filename).exists():
            return filename
        try:
            filename = os.environ["MORELIA_CONFIG"]
            if Path(filename).exists():
                return filename
        except KeyError:
            pass
        paths = [Path("pyproject.toml"), Path("~/.config/morelia/config.toml")]
        paths = [str(path) for path in paths if path.exists()]
        try:
            return paths[0]
        except IndexError:
            return None

    def get_matchers(self):
        return [MATCHERS[matcher] for matcher in self.__data["matchers"]]

    def get_writers(self):
        writers = []
        for writer_conf in self.__data.get("output", []):
            output_config = writer_conf["writer"]
            output = self.__create_output(output_config)
            format_config = writer_conf["formatter"]
            writers.append(self.__create_formatter(format_config, output))
        return writers

    def __create_output(self, config):
        output_type = config.pop("type")
        output_class = OUTPUTS[output_type]
        buffered = config.pop("buffered", output_class.default_buffered)
        output = output_class(**config)
        if buffered:
            output = Buffered(output)
        return output

    def __create_formatter(self, config, output):
        output_format = config.pop("format")
        formatter_class = FORMATTERS[output_format]
        return formatter_class(output=output, **config)

    def __getitem__(self, key):
        return self.__data[key]



# -*- coding: utf-8 -*-
#            __  __                _ _
#           |  \/  | ___  _ __ ___| (_) __ _
#           | |\/| |/ _ \| '__/ _ \ | |/ _` |
#           | |  | | (_) | | |  __/ | | (_| |
#           |_|  |_|\___/|_|  \___|_|_|\__,_|
#                             o        o     |  o
#                                 ,_       __|      ,
#                        |  |_|  /  |  |  /  |  |  / \_
#                         \/  |_/   |_/|_/\_/|_/|_/ \/
import re
import textwrap

from morelia.breadcrumbs import Breadcrumbs
from morelia.config import TOMLConfig
from morelia.exceptions import InvalidScenarioMatchingPattern, MissingStepError
from morelia.formatters import Writer
from morelia.grammar import (
    And,
    Background,
    But,
    Comment,
    Examples,
    Feature,
    Given,
    Row,
    Scenario,
    Step,
    Then,
    When,
)
from morelia.i18n import TRANSLATIONS
from morelia.visitors import TestVisitor


def execute_script(
    script_root,
    suite,
    scenario=".*",
    formatter=None,
    matchers=None,
    show_all_missing=True,
    config=None,
):
    try:
        scenario_re = re.compile(scenario)
    except re.error as e:
        raise InvalidScenarioMatchingPattern(
            'Invalid scenario matching regex "{}": {}'.format(scenario, e)
        )
    if config is None:
        config = TOMLConfig("default")
    matchers = __prepare_matchers(config, matchers, suite)
    wip = config["wip"]
    if not wip and show_all_missing:
        not_found = _find_missing_steps(script_root, matchers, scenario_re)
        message = "Cannot match steps:\n\n{}".format("".join(not_found))
        assert not_found == set(), message
    test_visitor = TestVisitor(suite, matchers, scenario_re)
    breadcrumbs = Breadcrumbs()
    test_visitor.register(breadcrumbs)
    __prepare_writers(config, formatter, test_visitor)
    try:
        script_root.accept(test_visitor)
    except Exception as exc:
        exc.__traceback__.tb_frame.f_locals
        tb = exc.__traceback__.tb_next
        while tb and not tb.tb_frame.f_locals.get("__tracebackhide__", False):
            tb = tb.tb_next
        if not tb:
            del tb
            raise
        exc.__traceback__ = tb.tb_next
        del tb
        raise exc from AssertionError(breadcrumbs)


def __prepare_matchers(config, matchers, suite):
    if matchers is None:
        matchers = config.get_matchers()
    matcher = _create_matchers_chain(suite, matchers)
    return matcher


def _create_matchers_chain(suite, matcher_classes):
    root_matcher = None
    for matcher_class in matcher_classes:
        matcher = matcher_class(suite)
        try:
            root_matcher.add_matcher(matcher)
        except AttributeError:
            root_matcher = matcher
    return root_matcher


def _find_missing_steps(feature, matcher, scenario_re):
    not_matched = {}  # "ordered dict/set" since 3.6 ;)
    for step in feature.get_all_steps():
        try:
            step.find_method(matcher)
        except MissingStepError as e:
            parent = step.parent
            if not isinstance(parent, Scenario) or scenario_re.match(
                step.parent.predicate
            ):
                not_matched[e.suggest] = True
    return not_matched.keys()


def __prepare_writers(config, formatter, test_visitor):
    writers = config.get_writers()
    if formatter is not None:
        writers.append(Writer(formatter))
    for writer in writers:
        test_visitor.register(writer)


class Parser:
    def __init__(self, language=None):
        self.__node_classes = [
            Feature,
            Background,
            Scenario,
            Given,
            When,
            Then,
            And,
            But,
            Row,
            Comment,
            Examples,
            Step,
        ]
        self.nodes = []
        if language is None:
            language = "en"
        self.__language = language
        self.__continuation_marker_re = re.compile(r"\\\s*$")

    def parse_features(self, prose):
        self.parse_feature(str(prose))
        feature = self.nodes[0]
        feature.filename = getattr(prose, "filename", "<stdin>")
        assert isinstance(feature, Feature), "Exactly one Feature per file"
        feature.enforce(
            any(isinstance(step, Scenario) for step in feature.steps),
            "Feature without Scenario(s)",
        )
        return self.nodes[0]

    def parse_feature(self, lines):
        self.__line_producer = LineSource(lines)
        self.__docstring_parser = DocStringParser(self.__line_producer)
        self.__language_parser = LanguageParser(default_language=self.__language)
        self.__labels_parser = LabelParser()
        try:
            while True:
                line = self.__line_producer.get_line()
                if line:
                    self.__parse_line(line)
        except StopIteration:
            pass
        return self.nodes

    def __parse_line(self, line):
        if self.__language_parser.parse(line):
            self.__language = self.__language_parser.language
            return

        if self.__labels_parser.parse(line):
            return

        if self.__docstring_parser.parse(line):
            previous = self.nodes[-1]
            previous.payload = self.__docstring_parser.payload
            return

        if self.__parse_node(line):
            return

        if 0 < len(self.nodes):
            self.__append_to_previous_node(line)
        else:
            line_number = self.__line_producer.line_number
            s = Step(line, line_number=line_number)
            feature_name = TRANSLATIONS[self.__language_parser.language].get(
                "feature", "Feature"
            )
            feature_name = feature_name.replace("|", " or ")
            s.enforce(False, "feature files must start with a %s" % feature_name)

    def __parse_node(self, line):
        line_number = self.__line_producer.line_number
        folded_lines = self.__read_folded_lines(line)
        line = line.rstrip()
        source = line + folded_lines
        for node_class in self.__node_classes:
            if node_class.match(line, self.__language):
                labels = self.__labels_parser.pop_labels()
                node = node_class(
                    source=source,
                    line_number=line_number,
                    labels=labels,
                    predecessors=self.nodes,
                    language=self.__language,
                )
                self.nodes.append(node)
                return node

    def __read_folded_lines(self, line):
        folded_lines = [""]
        while self.__continuation_marker_re.search(line):
            line = self.__line_producer.get_line()
            folded_lines.append(line)
        return "\n".join(folded_lines)

    def __append_to_previous_node(self, line):
        previous = self.nodes[-1]
        previous.append_line(line)


class LabelParser:
    def __init__(self, labels_pattern=r"@\w+"):
        self._labels = []
        self._labels_re = re.compile(labels_pattern)
        self._labels_prefix_re = re.compile(r"^\s*@")

    def parse(self, line):
        """Parse labels.

        :param str line: line to parse
        :returns: True if line contains labels
        :side effects: sets self._labels to parsed labels
        """
        if self._labels_prefix_re.match(line):
            matches = self._labels_re.findall(line)
            if matches:
                self._labels.extend(matches)
                return True
        return False

    def pop_labels(self):
        """Return labels.

        :returns: labels
        :side effects: clears current labels
        """
        labels = [label.strip("@") for label in self._labels]
        self._labels = []
        return labels


class LanguageParser:
    def __init__(self, lang_pattern=r"^# language: (\w+)", default_language=None):
        if default_language is None:
            default_language = "en"
        self.__language = default_language
        self.__lang_re = re.compile(lang_pattern)

    def parse(self, line):
        """Parse language directive.

        :param str line: line to parse
        :returns: True if line contains language directive
        :side effects: sets self.language to parsed language
        """
        match = self.__lang_re.match(line)
        if match:
            self.__language = match.groups()[0]
            return True
        return False

    @property
    def language(self):
        return self.__language


class DocStringParser:
    def __init__(self, source, pattern=r'\s*"""\s*'):
        self.__source = source
        self.__docstring_re = re.compile(pattern)
        self.__payload = []

    def parse(self, line):
        """Parse docstring payload.

        :param str line: first line to parse
        :returns: True if docstring parsed
        :side effects: sets self.payload to parsed docstring
        """
        match = self.__docstring_re.match(line)
        if match:
            start_line = line
            self.__payload = []
            line = self.__source.get_line()
            while line != start_line:
                self.__payload.append(line)
                line = self.__source.get_line()
            return True
        return False

    @property
    def payload(self):
        return textwrap.dedent("\n".join(self.__payload))


class LineSource:
    def __init__(self, text):
        self.__lines = iter(line for line in text.split("\n") if line)
        self.__line_number = 0

    def get_line(self):
        """Return next line.

        :returns: next line of text
        """
        self.__line_number += 1
        return next(self.__lines)

    @property
    def line_number(self):
        return self.__line_number



"""
Formatting output
=================

Morelia complies with Unix's `Rule of Silence` [#ROS]_ so when you hook it like this:

.. code-block:: python

    verify(filename, self)

and all tests passes it would say nothing:

.. code-block:: console

    $ python -m unittest test_acceptance
    .
    ----------------------------------------------------------------------
    Ran 1 test in 0.028s

    OK

(here's only information from test runner)

But when something went wrong it would complie with Unix's `Rule of Repair` [#ROR]_
and fail noisily:

.. code-block:: console

    F
    ======================================================================
    FAIL: test_addition (__main__.CalculatorTestCase)
    Addition feature.
    ----------------------------------------------------------------------
    AssertionError:
    Feature: Addition
        In order to avoid silly mistakes
        As a math idiot
        I want to be told the sum of two numbers
    Scenario: Add two numbers
        Then the result should be "120" on the screen

    The above exception was the direct cause of the following exception:

    Traceback (most recent call last):
      File "./test_acceptance.py", line 31, in test_addition
        verify(filename, self)
      File "(...)/morelia/__init__.py", line 119, in verify
        execute_script(feature, suite, scenario=scenario, config=conf)
      File "(...)/morelia/parser.py", line 75, in execute_script
        raise exc from AssertionError(breadcrumbs)
      File "./test_acceptance.py", line 46, in step_the_result_should_be_on_the_screen
        self.assertEqual(int(number), self.calculator.get_result())
    AssertionError: 120 != 121

    ----------------------------------------------------------------------
    Ran 1 test in 0.020s

    FAILED (failures=1)

Verbosity
---------

In Behaviour Driven Development participate both programmers and non-programmers
and the latter are often used to programs which report all the time what is going on.
So to make Morelia a little more verbose you can configure custom output in pyproject.toml file.
E.g.

.. code-block:: toml

    [tool.morelia.default.output]
    formatter.format="text"
    writer.type="terminal"

.. code-block:: python

    verify(filename, self)

.. code-block:: console

    Feature: Addition
        In order to avoid silly mistakes
        As a math idiot
        I want to be told the sum of two numbers
    Scenario: Add two numbers
        Given I have powered calculator on                       # pass  0.000s
        When I enter "50" into the calculator                    # pass  0.000s
        And I enter "70" into the calculator                     # pass  0.000s
        And I press add                                          # pass  0.001s
        Then the result should be "120" on the screen            # pass  0.001s
    Scenario: Subsequent additions
        Given I have powered calculator on                       # pass  0.000s
        When I enter "50" into the calculator                    # pass  0.000s
        And I enter "70" into the calculator                     # pass  0.000s
        And I press add                                          # pass  0.001s
        And I enter "20" into the calculator                     # pass  0.000s
        And I press add                                          # pass  0.001s
        Then the result should be "140" on the screen            # pass  0.001s
    .
    ----------------------------------------------------------------------
    Ran 1 test in 0.027s

    OK

Formatter Classes
-----------------

"""

import sys
import time
from abc import ABC, abstractmethod
from io import StringIO

import requests

from morelia.grammar import Feature, Step
from morelia.visitors import VisitorObserver

colors = {
    "normal": "\x1b[30m",
    "fail": "\x1b[31m",
    "error": "\x1b[31m",
    "pass": "\x1b[32m",
    "reset": "\x1b[0m",
}


class PlainTextFormatter:
    def __init__(self, stream=None):
        self._stream = stream if stream is not None else sys.stderr

    def output(self, node, line, status, duration):
        if isinstance(node, Feature):
            self._stream.write("\n")
        if isinstance(node, Step):
            status = status.lower()
            text = "{:<60} # {:<5} {:.3f}s\n".format(line.strip("\n"), status, duration)
        else:
            text = "%s\n" % line.strip("\n")
        self._stream.write(text)
        self._stream.flush()


class ColorTextFormatter(PlainTextFormatter):
    def output(self, node, line, status, duration):
        if isinstance(node, Feature):
            self._stream.write("\n")
        if isinstance(node, Step):
            status = status.lower()
            text = "{}{:<60} # {:.3f}s{}\n".format(
                colors[status], line.strip("\n"), duration, colors["reset"]
            )
        else:
            text = "%s\n" % line.strip("\n")
        self._stream.write(text)
        self._stream.flush()


class Writer(VisitorObserver):
    def __init__(self, formatter):
        self.__formatter = formatter

    def step_started(self, node):
        self.status = "pass"
        self.start_time = time.time()

    def step_failed(self, node):
        self.status = "fail"

    def step_errored(self, node):
        self.status = "error"

    def step_finished(self, node):
        duration = time.time() - self.start_time
        self.__formatter.output(node, node.interpolated_source(), self.status, duration)

    def node_started(self, node):
        self.__formatter.output(node, node.interpolated_source(), "", 0)


class IOutput(ABC):
    @abstractmethod
    def write(self, text):  # pragma: nocover
        pass

    @abstractmethod
    def close(self):  # pragma: nocover
        pass


class FileOutput(IOutput):
    default_buffered = True

    def __init__(self, path, open_func=open):
        self._path = path
        self.__open = open_func

    def __eq__(self, other):
        return (self.__class__ == other.__class__) and (self._path == other._path)

    def write(self, text):
        file = self._get_file()
        file.write(text)
        file.flush()

    def close(self):
        self._get_file().close()

    def _get_file(self):
        try:
            return self.__file
        except AttributeError:
            self.__file = self.__open(self._path, "w")
            return self.__file


class TerminalOutput(FileOutput):
    default_buffered = False

    def __init__(self, dest="stderr"):
        self._path = dest

    def _get_file(self):
        if self._path == "stdout":
            return sys.stdout
        else:
            return sys.stderr

    def close(self):
        pass


class RemoteOutput(IOutput):
    default_buffered = True

    def __init__(self, url, transport=requests):
        self._url = url
        self.__transport = transport

    def __eq__(self, other):
        return (self.__class__ == other.__class__) and (self._url == other._url)

    def write(self, text):
        self.__transport.post(self._url, text)

    def close(self):
        pass


class Buffered(IOutput):
    def __init__(self, wrapped: IOutput):
        self._wrapped = wrapped
        self.__buffer = StringIO()

    def __eq__(self, other):
        return (self.__class__ == other.__class__) and (self._wrapped == other._wrapped)

    def write(self, text):
        self.__buffer.write(text)

    def close(self):
        self._wrapped.write(self.__buffer.getvalue())
        self.__buffer.close()
        self._wrapped.close()


class TextFormat(VisitorObserver):
    def __init__(self, output: IOutput, color=False):
        self._color = color
        self._output = output

    def __eq__(self, other):
        return (
            (self.__class__ == other.__class__)
            and (self._color == other._color)
            and (self._output == other._output)
        )

    def step_started(self, node):
        self.status = "pass"
        self.start_time = time.time()

    def step_failed(self, node):
        self.status = "fail"

    def step_errored(self, node):
        self.status = "error"

    def step_finished(self, node):
        duration = time.time() - self.start_time
        line = node.interpolated_source()
        if self._color:
            text = "{}{:<60} # {:.3f}s{}\n".format(
                colors[self.status], line.strip("\n"), duration, colors["reset"]
            )
        else:
            text = "{:<60} # {:<5} {:.3f}s\n".format(
                line.strip("\n"), self.status, duration
            )
        self._output.write(text)

    def feature_started(self, node):
        self._output.write("\n")

    def node_started(self, node):
        line = node.interpolated_source()
        text = line.strip("\n") + "\n"
        self._output.write(text)

    def verify_finished(self):
        self._output.close()



# -*- coding: utf-8 -*-
import sys
from pathlib import Path

import requests

from morelia.config import TOMLConfig
from morelia.formatters import ColorTextFormatter, PlainTextFormatter
from morelia.parser import Parser, execute_script

__version__ = "0.9.0"


def has_color_support():
    """Check if color in terminal is supported."""
    return sys.platform != "win32"  # pragma: nocover


def run(
    filename,
    suite,
    as_str=None,
    scenario=r".*",
    verbose=False,
    show_all_missing=True,
    **kwargs
):  # NOQA
    """Parse file and run tests on given suite.

    :param str filename: file name
    :param unittest.TestCase suite: TestCase instance
    :param string as_str: None to use file or a string containing the feature to parse
    :param string scenario: a regex pattern to match the scenario to run
    :param boolean verbose: be verbose
    :param boolean show_all_missing: show all missing steps
    """
    formatter = kwargs.get("formatter", None)
    if verbose and not formatter:
        if has_color_support():
            formatter = ColorTextFormatter()
        else:
            formatter = PlainTextFormatter()
        kwargs["formatter"] = formatter
    kwargs["show_all_missing"] = show_all_missing
    if as_str is None:
        source = File(filename)
    else:
        source = Text(as_str, filename)
    feature = Parser().parse_features(source)
    return execute_script(feature, suite, scenario=scenario, **kwargs)


def verify(script, suite, scenario: str = ".*", config: str = "default") -> None:
    """Verifies script with steps from suite.

    :param script: feature script
    :param suite: object with steps defined
    :param str scenario: regex pattern for selecting single scenarios
    :param str config: section from configuration to apply

    Script can be passed directly to verify method as first argument.

    .. code-block:: python

        >>> from morelia import verify
        >>> verify(
        ...    \"""
        ...    Feature: Addition
        ...    Scenario: Add two numbers
        ...      Given I have entered 50 into the calculator
        ...      And I have entered 70 into the calculator
        ...      When I press add
        ...      Then the result should be 120 on the screen
        ...    \""",
        ...    test_case_with_steps,
        )

    When given path to file with script morelia will read that file and verify:

    .. code-block:: python

        >>> verify('calculator.feature', test_case_with_steps)

    Similary url pointing to script can be given:

    .. code-block:: python

        >>> verify('http://example.com/calculator', test_case_with_steps)

    Two last invocations will work only for single line strings.
    If it starts with "http[s]://" it is considered an url.
    If it ends with ".feature" it is considered a file.

    To explicity mark parameter's type it can be wrapped in helper classes:

    - :py:func:`File`
    - :py:func:`Url`
    - :py:func:`Text`

        >>> from morelia import verify, File, Text, Url
        >>> verify(File('calculator.txt'), test_case_with_steps)
        >>> verify(Url('http://example.com/calculator.feature'), test_case_with_steps)
        >>> verify(
        ...    Text(\"""
        ...    Feature: Addition
        ...    Scenario: Add two numbers
        ...      Given I have entered 50 into the calculator
        ...      And I have entered 70 into the calculator
        ...      When I press add
        ...      Then the result should be 120 on the screen
        ...    \"""),
        ...    test_case_with_steps,
        )

    """
    conf = TOMLConfig(config)
    script = _coerce_type(script)
    feature = Parser().parse_features(script)
    execute_script(feature, suite, scenario=scenario, config=conf)


def _coerce_type(script):
    if isinstance(script, Source):
        return script
    script = str(script)
    if "\n" in script:
        return Text(script)
    elif script.startswith("http://") or script.startswith("https://"):
        return Url(script)
    elif script.endswith(".feature"):
        return File(script)
    else:
        return Text(script)


class Source:
    pass


class Text(Source):
    """Marks string as a text source.

    :param str text: text source
    :param str name: optional name show in tracebacks
    """

    def __init__(self, text, name="<stdin>"):
        self.filename = name
        self.__text = text

    def __str__(self):
        return self.__text


class File(Source):
    """Marks string as a file path."""

    def __init__(self, path):
        self.__path = Path(path)
        self.filename = str(path)

    def __str__(self):
        try:
            return self.__text
        except AttributeError:
            self.__text = self.__path.read_text()
            return self.__text


class Url(Source):
    """Marks string as an url endpoint."""

    def __init__(self, url):
        self.__url = url
        self.filename = url

    def __str__(self):
        try:
            return self.__text
        except AttributeError:
            self.__text = requests.get(self.__url).text
            return self.__text


__all__ = ("Parser", "run", "verify", "File", "Url")



# -*- coding: utf-8 -*-

# source: https://github.com/cucumber/gherkin/blob/master/lib/gherkin/i18n.json

TRANSLATIONS = {
    "en": {
        "name": "English",
        "native": "English",
        "feature": "Feature|Business Need|Ability",
        "background": "Background",
        "scenario": "Scenario Outline|Scenario Template|Scenario",
        "examples": "Examples|Scenarios",
        "given": "Given",
        "when": "When",
        "then": "Then",
        "and": "And",
        "but": "But",
    },
    "af": {
        "name": "Afrikaans",
        "native": "Afrikaans",
        "feature": "Funksie|Besigheid Behoefte|Vermoë",
        "background": "Agtergrond",
        "scenario": "Situasie Uiteensetting|Situasie",
        "examples": "Voorbeelde",
        "given": "Gegewe",
        "when": "Wanneer",
        "then": "Dan",
        "and": "En",
        "but": "Maar",
    },
    "ar": {
        "name": "Arabic",
        "native": "العربية",
        "feature": "خاصية",
        "background": "الخلفية",
        "scenario": "سيناريو مخطط|سيناريو",
        "examples": "امثلة",
        "given": "بفرض",
        "when": "متى|عندما",
        "then": "اذاً|ثم",
        "and": "و",
        "but": "لكن",
    },
    "bm": {
        "name": "Malay",
        "native": "Bahasa Melayu",
        "feature": "Fungsi",
        "background": "Latar Belakang",
        "scenario": "Template Senario|Template Situai|Template Keadaan|Menggariskan Senario|Senario|Situai|Keadaan",
        "examples": "Contoh",
        "given": "Diberi|Bagi",
        "when": "Apabila",
        "then": "Maka|Kemudian",
        "and": "Dan",
        "but": "Tetapi|Tapi",
    },
    "bg": {
        "name": "Bulgarian",
        "native": "български",
        "feature": "Функционалност",
        "background": "Предистория",
        "scenario": "Рамка на сценарий|Сценарий",
        "examples": "Примери",
        "given": "Дадено",
        "when": "Когато",
        "then": "То",
        "and": "И",
        "but": "Но",
    },
    "ca": {
        "name": "Catalan",
        "native": "català",
        "background": "Rerefons|Antecedents",
        "feature": "Característica|Funcionalitat",
        "scenario": "Esquema de l'escenari|Escenari",
        "examples": "Exemples",
        "given": "Donat|Donada|Atès|Atesa",
        "when": "Quan",
        "then": "Aleshores|Cal",
        "and": "I",
        "but": "Però",
    },
    "cy-GB": {
        "name": "Welsh",
        "native": "Cymraeg",
        "background": "Cefndir",
        "feature": "Arwedd",
        "scenario": "Scenario Amlinellol|Scenario",
        "examples": "Enghreifftiau",
        "given": "Anrhegedig a",
        "when": "Pryd",
        "then": "Yna",
        "and": "A",
        "but": "Ond",
    },
    "cs": {
        "name": "Czech",
        "native": "Česky",
        "feature": "Požadavek",
        "background": "Pozadí|Kontext",
        "scenario": "Náčrt Scénáře|Osnova scénáře|Scénář",
        "examples": "Příklady",
        "given": "Pokud|Za předpokladu",
        "when": "Když",
        "then": "Pak",
        "and": "A také|A",
        "but": "Ale",
    },
    "da": {
        "name": "Danish",
        "native": "dansk",
        "feature": "Egenskab",
        "background": "Baggrund",
        "scenario": "Abstrakt Scenario|Scenarie",
        "examples": "Eksempler",
        "given": "Givet",
        "when": "Når",
        "then": "Så",
        "and": "Og",
        "but": "Men",
    },
    "de": {
        "name": "German",
        "native": "Deutsch",
        "feature": "Funktionalität",
        "background": "Grundlage",
        "scenario": "Szenariogrundriss|Szenario",
        "examples": "Beispiele",
        "given": "Angenommen|Gegeben sei|Gegeben seien",
        "when": "Wenn",
        "then": "Dann",
        "and": "Und",
        "but": "Aber",
    },
    "el": {
        "name": "Greek",
        "native": "Ελληνικά",
        "feature": "Δυνατότητα|Λειτουργία",
        "background": "Υπόβαθρο",
        "scenario": "Περιγραφή Σεναρίου|Σενάριο",
        "examples": "Παραδείγματα|Σενάρια",
        "given": "Δεδομένου",
        "when": "Όταν",
        "then": "Τότε",
        "and": "Και",
        "but": "Αλλά",
    },
    "en-au": {
        "name": "Australian",
        "native": "Australian",
        "feature": "Pretty much",
        "background": "First off",
        "scenario": "Reckon it's like|Awww, look mate",
        "examples": "You'll wanna",
        "given": "Y'know",
        "when": "It's just unbelievable",
        "then": "But at the end of the day I reckon",
        "and": "Too right",
        "but": "Yeah nah",
    },
    "en-Scouse": {
        "name": "Scouse",
        "native": "Scouse",
        "feature": "Feature",
        "background": "Dis is what went down",
        "scenario": "Wharrimean is|The thing of it is",
        "examples": "Examples",
        "given": "Givun|Youse know when youse got",
        "when": "Wun|Youse know like when",
        "then": "Dun|Den youse gotta",
        "and": "An",
        "but": "Buh",
    },
    "en-tx": {
        "name": "Texan",
        "native": "Texan",
        "feature": "Feature",
        "background": "Background",
        "scenario": "All y'all|Scenario",
        "examples": "Examples",
        "given": "Given y'all",
        "when": "When y'all",
        "then": "Then y'all",
        "and": "And y'all",
        "but": "But y'all",
    },
    "eo": {
        "name": "Esperanto",
        "native": "Esperanto",
        "feature": "Trajto",
        "background": "Fono",
        "scenario": "Konturo de la scenaro|Scenaro",
        "examples": "Ekzemploj",
        "given": "Donitaĵo",
        "when": "Se",
        "then": "Do",
        "and": "Kaj",
        "but": "Sed",
    },
    "es": {
        "name": "Spanish",
        "native": "español",
        "background": "Antecedentes",
        "feature": "Característica",
        "scenario": "Esquema del escenario|Escenario",
        "examples": "Ejemplos",
        "given": "Dado|Dada|Dados|Dadas",
        "when": "Cuando",
        "then": "Entonces",
        "and": "Y",
        "but": "Pero",
    },
    "et": {
        "name": "Estonian",
        "native": "eesti keel",
        "feature": "Omadus",
        "background": "Taust",
        "scenario": "Raamstsenaarium|Stsenaarium",
        "examples": "Juhtumid",
        "given": "Eeldades",
        "when": "Kui",
        "then": "Siis",
        "and": "Ja",
        "but": "Kuid",
    },
    "fa": {
        "name": "Persian",
        "native": "فارسی",
        "feature": "وِیژگی",
        "background": "زمینه",
        "scenario": "الگوی سناریو|سناریو",
        "examples": "نمونه ها",
        "given": "با فرض",
        "when": "هنگامی",
        "then": "آنگاه",
        "and": "و",
        "but": "اما",
    },
    "fi": {
        "name": "Finnish",
        "native": "suomi",
        "feature": "Ominaisuus",
        "background": "Tausta",
        "scenario": "Tapausaihio|Tapaus",
        "examples": "Tapaukset",
        "given": "Oletetaan",
        "when": "Kun",
        "then": "Niin",
        "and": "Ja",
        "but": "Mutta",
    },
    "fr": {
        "name": "French",
        "native": "français",
        "feature": "Fonctionnalité",
        "background": "Contexte",
        "scenario": "Plan du scénario|Plan du Scénario|Scénario",
        "examples": "Exemples",
        "given": "Soit|Etant donné|Etant donnée|Etant donnés|Etant données|Étant donné|Étant donnée|Étant donnés|Étant données",
        "when": "Quand|Lorsque|Lorsqu'<",
        "then": "Alors",
        "and": "Et",
        "but": "Mais",
    },
    "gl": {
        "name": "Galician",
        "native": "galego",
        "background": "Contexto",
        "feature": "Característica",
        "scenario": "Esbozo do escenario|Escenario",
        "examples": "Exemplos",
        "given": "Dado|Dada|Dados|Dadas",
        "when": "Cando",
        "then": "Entón|Logo",
        "and": "E",
        "but": "Mais|Pero",
    },
    "he": {
        "name": "Hebrew",
        "native": "עברית",
        "feature": "תכונה",
        "background": "רקע",
        "scenario": "תבנית תרחיש|תרחיש",
        "examples": "דוגמאות",
        "given": "בהינתן",
        "when": "כאשר",
        "then": "אז|אזי",
        "and": "וגם",
        "but": "אבל",
    },
    "hi": {
        "name": "Hindi",
        "native": "हिंदी",
        "feature": "रूप लेख",
        "background": "पृष्ठभूमि",
        "scenario": "परिदृश्य रूपरेखा|परिदृश्य",
        "examples": "उदाहरण",
        "given": "अगर|यदि|चूंकि",
        "when": "जब|कदा",
        "then": "तब|तदा",
        "and": "और|तथा",
        "but": "पर|परन्तु|किन्तु",
    },
    "hr": {
        "name": "Croatian",
        "native": "hrvatski",
        "feature": "Osobina|Mogućnost|Mogucnost",
        "background": "Pozadina",
        "scenario": "Skica|Koncept|Scenarij",
        "examples": "Primjeri|Scenariji",
        "given": "Zadan|Zadani|Zadano",
        "when": "Kada|Kad",
        "then": "Onda",
        "and": "I",
        "but": "Ali",
    },
    "ht": {
        "name": "Creole",
        "native": "kreyòl",
        "feature": "Karakteristik|Mak|Fonksyonalite",
        "background": "Kontèks|Istorik",
        "scenario": "Plan senaryo|Plan Senaryo|Senaryo deskripsyon|Senaryo Deskripsyon|Dyagram senaryo|Dyagram Senaryo|Senaryo",
        "examples": "Egzanp",
        "given": "Sipoze|Sipoze ke|Sipoze Ke",
        "when": "Lè|Le",
        "then": "Lè sa a|Le sa a",
        "and": "Ak|Epi|E",
        "but": "Men",
    },
    "hu": {
        "name": "Hungarian",
        "native": "magyar",
        "feature": "Jellemző",
        "background": "Háttér",
        "scenario": "Forgatókönyv vázlat|Forgatókönyv",
        "examples": "Példák",
        "given": "Amennyiben|Adott",
        "when": "Majd|Ha|Amikor",
        "then": "Akkor",
        "and": "És",
        "but": "De",
    },
    "id": {
        "name": "Indonesian",
        "native": "Bahasa Indonesia",
        "feature": "Fitur",
        "background": "Dasar",
        "scenario": "Skenario konsep|Skenario",
        "examples": "Contoh",
        "given": "Dengan",
        "when": "Ketika",
        "then": "Maka",
        "and": "Dan",
        "but": "Tapi",
    },
    "is": {
        "name": "Icelandic",
        "native": "Íslenska",
        "feature": "Eiginleiki",
        "background": "Bakgrunnur",
        "scenario": "Lýsing Atburðarásar|Lýsing Dæma|Atburðarás",
        "examples": "Dæmi|Atburðarásir",
        "given": "Ef",
        "when": "Þegar",
        "then": "Þá",
        "and": "Og",
        "but": "En",
    },
    "it": {
        "name": "Italian",
        "native": "italiano",
        "feature": "Funzionalità",
        "background": "Contesto",
        "scenario": "Schema dello scenario|Scenario",
        "examples": "Esempi",
        "given": "Dato|Data|Dati|Date",
        "when": "Quando",
        "then": "Allora",
        "and": "E",
        "but": "Ma",
    },
    "ja": {
        "name": "Japanese",
        "native": "日本語",
        "feature": "フィーチャ|機能",
        "background": "背景",
        "scenario": "シナリオアウトライン|シナリオテンプレート|テンプレ|シナリオテンプレ|シナリオ",
        "examples": "例|サンプル",
        "given": "前提<",
        "when": "もし<",
        "then": "ならば<",
        "and": "かつ<",
        "but": "しかし<|但し<|ただし<",
    },
    "jv": {
        "name": "Javanese",
        "native": "Basa Jawa",
        "feature": "Fitur",
        "background": "Dasar",
        "scenario": "Konsep skenario|Skenario",
        "examples": "Conto|Contone",
        "given": "Nalika|Nalikaning",
        "when": "Manawa|Menawa",
        "then": "Njuk|Banjur",
        "and": "Lan",
        "but": "Tapi|Nanging|Ananging",
    },
    "kn": {
        "name": "Kannada",
        "native": "ಕನ್ನಡ",
        "background": "ಹಿನ್ನೆಲೆ",
        "feature": "ಹೆಚ್ಚಳ",
        "scenario": "ವಿವರಣೆ|ಕಥಾಸಾರಾಂಶ",
        "examples": "ಉದಾಹರಣೆಗಳು",
        "given": "ನೀಡಿದ",
        "when": "ಸ್ಥಿತಿಯನ್ನು",
        "then": "ನಂತರ",
        "and": "ಮತ್ತು",
        "but": "ಆದರೆ",
    },
    "ko": {
        "name": "Korean",
        "native": "한국어",
        "background": "배경",
        "feature": "기능",
        "scenario": "시나리오 개요|시나리오",
        "examples": "예",
        "given": "조건<|먼저<",
        "when": "만일<|만약<",
        "then": "그러면<",
        "and": "그리고<",
        "but": "하지만<|단<",
    },
    "lt": {
        "name": "Lithuanian",
        "native": "lietuvių kalba",
        "feature": "Savybė",
        "background": "Kontekstas",
        "scenario": "Scenarijaus šablonas|Scenarijus",
        "examples": "Pavyzdžiai|Scenarijai|Variantai",
        "given": "Duota",
        "when": "Kai",
        "then": "Tada",
        "and": "Ir",
        "but": "Bet",
    },
    "lu": {
        "name": "Luxemburgish",
        "native": "Lëtzebuergesch",
        "feature": "Funktionalitéit",
        "background": "Hannergrond",
        "scenario": "Plang vum Szenario|Szenario",
        "examples": "Beispiller",
        "given": "ugeholl",
        "when": "wann",
        "then": "dann",
        "and": "an|a",
        "but": "awer|mä",
    },
    "lv": {
        "name": "Latvian",
        "native": "latviešu",
        "feature": "Funkcionalitāte|Fīča",
        "background": "Konteksts|Situācija",
        "scenario": "Scenārijs pēc parauga|Scenārijs",
        "examples": "Piemēri|Paraugs",
        "given": "Kad",
        "when": "Ja",
        "then": "Tad",
        "and": "Un",
        "but": "Bet",
    },
    "nl": {
        "name": "Dutch",
        "native": "Nederlands",
        "feature": "Functionaliteit",
        "background": "Achtergrond",
        "scenario": "Abstract Scenario|Scenario",
        "examples": "Voorbeelden",
        "given": "Gegeven|Stel",
        "when": "Als",
        "then": "Dan",
        "and": "En",
        "but": "Maar",
    },
    "no": {
        "name": "Norwegian",
        "native": "norsk",
        "feature": "Egenskap",
        "background": "Bakgrunn",
        "scenario": "Scenariomal|Abstrakt Scenario|Scenario",
        "examples": "Eksempler",
        "given": "Gitt",
        "when": "Når",
        "then": "Så",
        "and": "Og",
        "but": "Men",
    },
    "pa": {
        "name": "Panjabi",
        "native": "ਪੰਜਾਬੀ",
        "feature": "ਖਾਸੀਅਤ|ਮੁਹਾਂਦਰਾ|ਨਕਸ਼ ਨੁਹਾਰ",
        "background": "ਪਿਛੋਕੜ",
        "scenario": "ਪਟਕਥਾ ਢਾਂਚਾ|ਪਟਕਥਾ ਰੂਪ ਰੇਖਾ|ਪਟਕਥਾ",
        "examples": "ਉਦਾਹਰਨਾਂ",
        "given": "ਜੇਕਰ|ਜਿਵੇਂ ਕਿ",
        "when": "ਜਦੋਂ",
        "then": "ਤਦ",
        "and": "ਅਤੇ",
        "but": "ਪਰ",
    },
    "pl": {
        "name": "Polish",
        "native": "polski",
        "feature": "Właściwość|Funkcja|Aspekt|Potrzeba biznesowa",
        "background": "Założenia",
        "scenario": "Szablon scenariusza|Scenariusz",
        "examples": "Przykłady",
        "given": "Zakładając(?:,? że)?|Mając",
        "when": "Jeżeli|Jeśli|Gdy|Kiedy",
        "then": "Wtedy",
        "and": "Oraz|I",
        "but": "Ale",
    },
    "pt": {
        "name": "Portuguese",
        "native": "português",
        "background": "Contexto|Cenário de Fundo|Cenario de Fundo|Fundo",
        "feature": "Funcionalidade|Característica|Caracteristica",
        "scenario": "Esquema do Cenário|Esquema do Cenario|Delineação do Cenário|Delineacao do Cenario|Cenário|Cenario",
        "examples": "Exemplos|Cenários|Cenarios",
        "given": "Dado|Dada|Dados|Dadas",
        "when": "Quando",
        "then": "Então|Entao",
        "and": "E",
        "but": "Mas",
    },
    "ro": {
        "name": "Romanian",
        "native": "română",
        "background": "Context",
        "feature": "Functionalitate|Funcționalitate|Funcţionalitate",
        "scenario": "Structura scenariu|Structură scenariu|Scenariu",
        "examples": "Exemple",
        "given": "Date fiind|Dat fiind|Dati fiind|Dați fiind|Daţi fiind",
        "when": "Cand|Când",
        "then": "Atunci",
        "and": "Si|Și|Şi",
        "but": "Dar",
    },
    "ru": {
        "name": "Russian",
        "native": "русский",
        "feature": "Функция|Функционал|Свойство",
        "background": "Предыстория|Контекст",
        "scenario": "Структура сценария|Сценарий",
        "examples": "Примеры",
        "given": "Допустим|Дано|Пусть",
        "when": "Если|Когда",
        "then": "То|Тогда",
        "and": "И|К тому же|Также",
        "but": "Но|А",
    },
    "sv": {
        "name": "Swedish",
        "native": "Svenska",
        "feature": "Egenskap",
        "background": "Bakgrund",
        "scenario": "Abstrakt Scenario|Scenariomall|Scenario",
        "examples": "Exempel",
        "given": "Givet",
        "when": "När",
        "then": "Så",
        "and": "Och",
        "but": "Men",
    },
    "sk": {
        "name": "Slovak",
        "native": "Slovensky",
        "feature": "Požiadavka|Funkcia|Vlastnosť",
        "background": "Pozadie",
        "scenario": "Náčrt Scenáru|Náčrt Scenára|Osnova Scenára|Scenár",
        "examples": "Príklady",
        "given": "Pokiaľ|Za predpokladu",
        "when": "Keď|Ak",
        "then": "Tak|Potom",
        "and": "A|A tiež|A taktiež|A zároveň",
        "but": "Ale",
    },
    "sl": {
        "name": "Slovenian",
        "native": "Slovenski",
        "feature": "Funkcionalnost|Funkcija|Možnosti|Moznosti|Lastnost|Značilnost",
        "background": "Kontekst|Osnova|Ozadje",
        "scenario": "Struktura scenarija|Skica|Koncept|Oris scenarija|Osnutek|Scenarij|Primer",
        "examples": "Primeri|Scenariji",
        "given": "Dano|Podano|Zaradi|Privzeto",
        "when": "Ko|Ce|Če|Kadar",
        "then": "Nato|Potem|Takrat",
        "and": "In|Ter",
        "but": "Toda|Ampak|Vendar",
    },
    "sr-Latn": {
        "name": "Serbian (Latin)",
        "native": "Srpski (Latinica)",
        "feature": "Funkcionalnost|Mogućnost|Mogucnost|Osobina",
        "background": "Kontekst|Osnova|Pozadina",
        "scenario": "Struktura scenarija|Skica|Koncept|Scenario|Primer",
        "examples": "Primeri|Scenariji",
        "given": "Zadato|Zadate|Zatati",
        "when": "Kada|Kad",
        "then": "Onda",
        "and": "I",
        "but": "Ali",
    },
    "sr-Cyrl": {
        "name": "Serbian",
        "native": "Српски",
        "feature": "Функционалност|Могућност|Особина",
        "background": "Контекст|Основа|Позадина",
        "scenario": "Структура сценарија|Скица|Концепт|Сценарио|Пример",
        "examples": "Примери|Сценарији",
        "given": "Задато|Задате|Задати",
        "when": "Када|Кад",
        "then": "Онда",
        "and": "И",
        "but": "Али",
    },
    "tl": {
        "name": "Telugu",
        "native": "తెలుగు",
        "feature": "గుణము",
        "background": "నేపథ్యం",
        "scenario": "కథనం|సన్నివేశం",
        "examples": "ఉదాహరణలు",
        "given": "చెప్పబడినది",
        "when": "ఈ పరిస్థితిలో",
        "then": "అప్పుడు",
        "and": "మరియు",
        "but": "కాని",
    },
    "th": {
        "name": "Thai",
        "native": "ไทย",
        "feature": "โครงหลัก|ความต้องการทางธุรกิจ|ความสามารถ",
        "background": "แนวคิด",
        "scenario": "สรุปเหตุการณ์|โครงสร้างของเหตุการณ์|เหตุการณ์",
        "examples": "ชุดของตัวอย่าง|ชุดของเหตุการณ์",
        "given": "กำหนดให้",
        "when": "เมื่อ",
        "then": "ดังนั้น",
        "and": "และ",
        "but": "แต่",
    },
    "tlh": {
        "name": "Klingon",
        "native": "tlhIngan",
        "feature": "Qap|Qu'meH 'ut|perbogh|poQbogh malja'|laH",
        "background": "mo'",
        "scenario": "lut chovnatlh|lut",
        "examples": "ghantoH|lutmey",
        "given": "ghu' noblu'|DaH ghu' bejlu'",
        "when": "qaSDI'",
        "then": "vaj",
        "and": "'ej|latlh",
        "but": "'ach|'a",
    },
    "tr": {
        "name": "Turkish",
        "native": "Türkçe",
        "feature": "Özellik",
        "background": "Geçmiş",
        "scenario": "Senaryo taslağı|Senaryo",
        "examples": "Örnekler",
        "given": "Diyelim ki",
        "when": "Eğer ki",
        "then": "O zaman",
        "and": "Ve",
        "but": "Fakat|Ama",
    },
    "tt": {
        "name": "Tatar",
        "native": "Татарча",
        "feature": "Мөмкинлек|Үзенчәлеклелек",
        "background": "Кереш",
        "scenario": "Сценарийның төзелеше|Сценарий",
        "examples": "Үрнәкләр|Мисаллар",
        "given": "Әйтик",
        "when": "Әгәр",
        "then": "Нәтиҗәдә",
        "and": "Һәм|Вә",
        "but": "Ләкин|Әмма",
    },
    "uk": {
        "name": "Ukrainian",
        "native": "Українська",
        "feature": "Функціонал",
        "background": "Передумова",
        "scenario": "Структура сценарію|Сценарій",
        "examples": "Приклади",
        "given": "Припустимо|Припустимо, що|Нехай|Дано",
        "when": "Якщо|Коли",
        "then": "То|Тоді",
        "and": "І|А також|Та",
        "but": "Але",
    },
    "uz": {
        "name": "Uzbek",
        "native": "Узбекча",
        "feature": "Функционал",
        "background": "Тарих",
        "scenario": "Сценарий структураси|Сценарий",
        "examples": "Мисоллар",
        "given": "Агар",
        "when": "Агар",
        "then": "Унда",
        "and": "Ва",
        "but": "Лекин|Бирок|Аммо",
    },
    "vi": {
        "name": "Vietnamese",
        "native": "Tiếng Việt",
        "feature": "Tính năng",
        "background": "Bối cảnh",
        "scenario": "Khung tình huống|Khung kịch bản|Tình huống|Kịch bản",
        "examples": "Dữ liệu",
        "given": "Biết|Cho",
        "when": "Khi",
        "then": "Thì",
        "and": "Và",
        "but": "Nhưng",
    },
    "zh-CN": {
        "name": "Chinese simplified",
        "native": "简体中文",
        "feature": "功能",
        "background": "背景",
        "scenario": "场景大纲|剧本大纲|场景|剧本",
        "examples": "例子",
        "given": "假如<|假设<|假定<",
        "when": "当<",
        "then": "那么<",
        "and": "而且<|并且<|同时<",
        "but": "但是<",
    },
    "zh-TW": {
        "name": "Chinese traditional",
        "native": "繁體中文",
        "feature": "功能",
        "background": "背景",
        "scenario": "場景大綱|劇本大綱|場景|劇本",
        "examples": "例子",
        "given": "假如<|假設<|假定<",
        "when": "當<",
        "then": "那麼<",
        "and": "而且<|並且<|同時<",
        "but": "但是<",
    },
    "ur": {
        "name": "Urdu",
        "native": "اردو",
        "feature": "صلاحیت|کاروبار کی ضرورت|خصوصیت",
        "background": "پس منظر",
        "scenario": "منظر نامے کا خاکہ|منظرنامہ",
        "examples": "مثالیں",
        "given": "اگر|بالفرض|فرض کیا",
        "when": "جب",
        "then": "پھر|تب",
        "and": "اور",
        "but": "لیکن",
    },
}



import copy
import itertools
import re
from abc import ABC, abstractmethod
from typing import Iterable, Type

from morelia.exceptions import MissingStepError
from morelia.i18n import TRANSLATIONS

PLACEHOLDER_RE = re.compile(r"\<(\w+)\>")


class Visitor(ABC):  # pragma: nocover
    @abstractmethod
    def visit_feature(self, node: "Feature", children: Iterable["Node"]) -> None:
        pass

    @abstractmethod
    def visit_scenario(self, node: "Scenario", children: Iterable["Node"]) -> None:
        pass

    @abstractmethod
    def visit_step(self, node: "Step", children: Iterable["Node"]) -> None:
        pass

    @abstractmethod
    def visit_background(self, node: "Background") -> None:
        pass

    @abstractmethod
    def visit_row(self, node: "Row") -> None:
        pass

    @abstractmethod
    def visit_examples(self, node: "Examples", children: Iterable["Node"]) -> None:
        pass

    @abstractmethod
    def visit_comment(self, node: "Comment") -> None:
        pass


class Node(ABC):
    allowed_parents = ()  # type: Iterable[Type[Node]]

    def __init__(
        self, source="", line_number=0, language="en", labels=None, predecessors=[]
    ):
        self.__source = source
        self.__line_number = line_number
        self.__language = language
        self.__labels = labels if labels is not None else []
        self.steps = []
        self.parent = None
        self.__predicate = self.__extract_predicate()
        self.parent = self.__find_parent(predecessors)
        self.__connect_to_parent()
        self._validate_predicate()

    def __connect_to_parent(self):
        if self.parent:
            self.parent.add_child(self)

    def __find_parent(self, predecessors):
        allowed_parents = self.allowed_parents
        if not allowed_parents and predecessors:
            self.enforce(False, "Only one Feature per file")
        for step in predecessors[::-1]:
            if isinstance(step, allowed_parents):
                return step
        return None

    def __extract_predicate(self):
        node_re = self.__get_compiled_pattern(self.__language)
        return node_re.sub("", self.source).strip()

    @classmethod
    def match(cls, line, language):
        node_re = cls.__get_compiled_pattern(language)
        return node_re.match(line)

    @classmethod
    def __get_compiled_pattern(cls, language, __memo={}):
        try:
            return __memo[cls, language]
        except KeyError:
            pattern = cls._get_pattern(language)
            node_re = __memo[cls, language] = re.compile(pattern)
            return node_re

    @classmethod
    def _get_pattern(cls, language):
        class_name = cls.__name__
        name = class_name.lower()
        name = TRANSLATIONS[language].get(name, class_name)
        return r"^\s*({name}):?(\s+|$)".format(name=name)

    @property
    def source(self):
        return self.__source

    @property
    def line_number(self):
        return self.__line_number

    @property
    def predicate(self):
        return self.__predicate

    def append_line(self, line):
        self.__source += "\n" + line
        self.__predicate = (self.__predicate + "\n" + line.strip()).strip()
        self._validate_predicate()

    def _validate_predicate(self):
        return  # looks good! (-:

    def get_labels(self):
        labels = self.__labels[:]
        if self.parent:
            labels.extend(self.parent.get_labels())
        return labels

    def get_all_steps(self):
        return itertools.chain.from_iterable(
            child.get_all_steps() for child in self.steps
        )

    def add_child(self, child):
        self.steps.append(child)

    def enforce(self, condition, diagnostic):
        if not condition:
            text = ""
            offset = 1
            if self.parent:
                text = self.parent.source
                offset = 5
            text += self.source
            text = text.replace("\n\n", "\n").replace("\n", "\n\t")
            raise SyntaxError(
                diagnostic, (self.get_filename(), self.line_number, offset, text)
            )

    def get_filename(self):
        try:
            return self.parent.get_filename() if self.parent else self.filename
        except AttributeError:
            return None

    def interpolated_source(self):
        return self.source + "\n"

    def format_fault(self, diagnostic):
        parent_reconstruction = ""
        if self.parent:
            parent_reconstruction = self.parent.source.strip("\n")
        args = (
            self.get_filename(),
            self.line_number,
            parent_reconstruction,
            self.source,
            diagnostic,
        )
        return '\n  File "%s", line %s, in %s\n %s\n%s' % args

    @abstractmethod
    def accept(self, visitor: Visitor) -> None:  # pragma: nocover
        pass


class Feature(Node):
    def accept(self, visitor: Visitor) -> None:
        visitor.visit_feature(self, self.steps)

    def prepend_steps(self, scenario):
        background = self.steps[0]
        try:
            background.prepend_steps(scenario)
        except AttributeError:
            pass


class Scenario(Node):
    allowed_parents = (Feature,)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.row_indices = [0]

    def accept(self, visitor: Visitor) -> None:
        self.parent.prepend_steps(self)
        self.enforce(
            0 < len(self.steps),
            "Scenario without step(s) - Step, Given, When, Then, And, or #",
        )
        schedule = self.permute_schedule()

        old_row_indices = self.row_indices
        try:
            for indices in schedule:
                self.row_indices = indices
                visitor.visit_scenario(self, self.steps)
        finally:
            self.row_indices = old_row_indices

    def permute_schedule(self):
        dims = self.count_Row_dimensions()
        return _permute_indices(dims)

    def count_Row_dimensions(self):
        return [step.rows_number for step in self.steps if isinstance(step, RowParent)]

    def get_all_steps(self):
        return (step for step in self.steps if isinstance(step, Step))


class Background(Node):
    allowed_parents = (Feature,)

    def accept(self, visitor: Visitor) -> None:
        visitor.visit_background(self)

    def prepend_steps(self, scenario):
        try:
            return scenario.background_steps
        except AttributeError:
            background_steps = []
            for step in self.steps:
                new_step = copy.copy(step)
                new_step.parent = scenario
                background_steps.append(new_step)
            scenario.steps = background_steps + scenario.steps
            scenario.background_steps = background_steps
            return background_steps

    def count_Row_dimensions(self):
        return [0]

    def get_all_steps(self):
        return (step for step in self.steps if isinstance(step, Step))


class RowParent(Node):
    @property
    def rows_number(self):
        rows_number = len(self.get_rows()) - 1  # do not count header
        return max(0, rows_number)

    def get_rows(self):
        return [step for step in self.steps if isinstance(step, Row)]


class Step(RowParent):
    allowed_parents = (Scenario, Background)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.payload = ""

    def accept(self, visitor: Visitor) -> None:
        visitor.visit_step(self, self.steps)

    def find_method(self, matcher):
        """Find method matching step.

        :param IStepMatcher matcher: object matching methods by given predicate
        :returns: (method, args, kwargs) tuple
        :rtype: tuple
        :raises MissingStepError: if method maching step not found
        """
        predicate = self.predicate
        augmented_predicate = self.__get_interpolated_predicate()
        method, args, kwargs = matcher.find(predicate, augmented_predicate)
        if method:
            return method, args, kwargs

        suggest, method_name, docstring = matcher.suggest(predicate)
        raise MissingStepError(predicate, suggest, method_name, docstring)

    def interpolated_source(self):
        augmented_predicate = self.__get_interpolated_predicate()
        return self.source.replace(self.predicate, augmented_predicate)

    def __get_interpolated_predicate(self):
        if self.parent is None:
            return self.predicate
        if self.__parent_has_no_rows():
            return self.predicate
        placeholders = PLACEHOLDER_RE.findall(self.predicate)
        if not placeholders:
            return self.predicate
        return self.__replace_placeholders_in_predicate(placeholders)

    def __parent_has_no_rows(self):
        dims = self.parent.count_Row_dimensions()
        return not any(dims)

    def __replace_placeholders_in_predicate(self, placeholders):
        copy = self.predicate[:]
        row_indices = self.parent.row_indices
        siblings = self.parent.steps
        for step_idx, row_idx in enumerate(row_indices):
            step = siblings[step_idx]
            table = step.get_rows()
            if len(table) > 1:
                header = table[0]
                body = table[1:]
                for column_idx, column_title in enumerate(header.values):
                    value = body[row_idx][column_idx]
                    copy = self.__replace_placeholders(
                        column_title, value, placeholders, copy
                    )
        return copy

    def __replace_placeholders(self, column_title, table_value, placeholders, copy):
        for placeholder in placeholders:
            if column_title == placeholder:
                return self.__replace_placeholder(copy, placeholder, table_value)
        return copy

    def __replace_placeholder(self, copy, placeholder, table_value):
        table_value = table_value.replace("\n", "\\n")
        return copy.replace(
            "<{placeholder}>".format(placeholder=placeholder), table_value
        )


class Given(Step):
    pass


class When(Step):
    pass


class Then(Step):
    pass


class And(Step):
    pass


class But(And):
    pass


class Examples(RowParent):
    allowed_parents = (Scenario,)

    def accept(self, visitor: Visitor) -> None:
        visitor.visit_examples(self, self.steps)


class Row(Node):
    allowed_parents = (Step, Examples)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._validate_predicate()

    def _validate_predicate(self):
        # TODO: validate that grandparent is not Background
        row = re.split(r" \|", re.sub(r"\|$", "", self.predicate))
        self.__values = [s.strip() for s in row]

    def __getitem__(self, column_idx):
        return self.__values[column_idx]

    @property
    def values(self):
        return self.__values

    def accept(self, visitor: Visitor) -> None:
        visitor.visit_row(self)

    @classmethod
    def _get_pattern(cls, language):
        return r"^\s*(\|):?\s+"


class Comment(Node):
    allowed_parents = (Node,)

    def accept(self, visitor: Visitor) -> None:
        visitor.visit_comment(self)

    @classmethod
    def _get_pattern(cls, language):
        return r"\s*(\#)"

    def _validate_predicate(self):
        self.enforce("\n" not in self.predicate, "linefeed in comment")


def _permute_indices(arr):
    product_args = list(_imap(arr))
    result = list(itertools.product(*product_args))
    return result
    #  tx to Chris Rebert, et al, on the Python newsgroup for curing my brainlock here!!


def _imap(*iterables):
    iterables = [iter(i) for i in iterables]
    while True:
        try:
            args = [next(i) for i in iterables]
            yield _special_range(*args)
        except StopIteration:
            return


def _special_range(n):
    return range(n) if n else [0]



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['morelia']

package_data = \
{'': ['*']}

install_requires = \
['parse>=1.12.0,<2.0.0', 'requests>=2.21,<3.0']

setup_kwargs = {
    'name': 'morelia',
    'version': '0.9.0',
    'description': 'for "Behavior Driven Development" (BDD) -- a client-facing scripting language to put the squeeze on all your features',
    'long_description': '#######\nMorelia\n#######\n\n.. image:: https://img.shields.io/pypi/wheel/Morelia.svg\n    :target: https://pypi.python.org/pypi/Morelia/\n    :alt: Wheel Status\n\n.. image:: https://img.shields.io/pypi/pyversions/Morelia.svg\n    :target: https://pypi.python.org/pypi/Morelia/\n    :alt: Python versions\n\n.. image:: https://img.shields.io/pypi/v/Morelia.svg\n    :target: https://pypi.python.org/pypi/Morelia/\n    :alt: Latest Version\n\n.. image:: https://img.shields.io/pypi/l/Morelia.svg\n    :target: https://pypi.python.org/pypi/Morelia/\n    :alt: License\n\n.. image:: https://travis-ci.org/kidosoft/Morelia.svg?branch=master\n    :target: https://travis-ci.org/kidosoft/Morelia\n    :alt: Build status\n\n.. image:: https://coveralls.io/repos/kidosoft/Morelia/badge.svg\n    :target: https://coveralls.io/r/kidosoft/Morelia\n    :alt: Coverage\n\n.. image:: https://readthedocs.org/projects/morelia/badge/?format=svg\n    :target: https://morelia.readthedocs.io\n    :alt: Documentation\n\n.. image:: https://pyup.io/repos/github/kidosoft/Morelia/shield.svg\n    :target: https://pyup.io/repos/github/kidosoft/Morelia/\n    :alt: Updates\n\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github.com/ambv/black\n\nMorelia is a Python Behavior Driven Development (BDD) library.\n\nBDD is an agile software development process that encourages\ncollaboration between developers, QA and business participants.\n\nTest scenarios written in natural language make BDD foundation.\nThey are comprehensible for non-technical participants who wrote them\nand unambiguous for developers and QA.\n\nMorelia makes it easy for developers to integrate BDD into their existing\nunittest frameworks.  It is easy to run under nose, pytest, tox, trial or integrate\nwith django, flask or any other python framework because no special code\nhave to be written.\n\nYou as developer are in charge of how tests are organized. No need to fit into\nrigid rules forced by some other BDD frameworks.\n\n**Mascot**:\n\n.. image:: http://www.naturfoto.cz/fotografie/ostatni/krajta-zelena-47784.jpg\n\nInstallation\n============\n\n.. code-block:: console\n\n    pip install morelia\n\nQuick usage guide\n=================\n\nWrite a feature description:\n\n.. code-block:: cucumber\n\n    # calculator.feature\n\n    Feature: Addition\n        In order to avoid silly mistakes\n        As a math idiot\n        I want to be told the sum of two numbers\n\n    Scenario: Add two numbers\n        Given I have powered calculator on\n        When I enter "50" into the calculator\n        And I enter "70" into the calculator\n        And I press add\n        Then the result should be "120" on the screen\n\n\nCreate standard Python\'s `unittest` and hook Morelia into it:\n\n.. code-block:: python\n\n    # test_acceptance.py\n\n    import unittest\n\n    from morelia import verify\n\n\n    class CalculatorTestCase(unittest.TestCase):\n    \n        def test_addition(self):\n            """ Addition feature """\n            verify(\'calculator.feature\', self)\n\nRun test with your favourite runner: unittest, nose, py.test, trial. You name it!\n\n.. code-block:: console\n\n    $ python -m unittest -v test_acceptance  # or\n    $ pytest test_acceptance.py  # or\n    $ nosetests -v test_acceptance.py  # or\n    $ trial test_acceptance.py  # or\n    $ # django/pyramid/flask/(place for your favourite test runner)\n\nAnd you\'ll see which steps are missing:\n\n.. code-block:: python\n\n    F\n    ======================================================================\n    FAIL: test_addition (test_acceptance.CalculatorTestCase)\n    Addition feature.\n    ----------------------------------------------------------------------\n    Traceback (most recent call last):\n      File "(..)test_acceptance.py", line 31, in test_addition\n        verify(filename, self)\n      File "(..)/morelia/__init__.py", line 120, in verify\n        execute_script(feature, suite, scenario=scenario, config=conf)\n      File "(..)/morelia/parser.py", line 59, in execute_script\n        assert not_found == set(), message\n    AssertionError: Cannot match steps:\n\n        def step_I_have_powered_calculator_on(self):\n            r\'I have powered calculator on\'\n\n            raise NotImplementedError(\'I have powered calculator on\')\n\n        def step_I_enter_number_into_the_calculator(self, number):\n            r\'I enter "([^"]+)" into the calculator\'\n\n            raise NotImplementedError(\'I enter "50" into the calculator\')\n\n        def step_I_enter_number_into_the_calculator(self, number):\n            r\'I enter "([^"]+)" into the calculator\'\n\n            raise NotImplementedError(\'I enter "70" into the calculator\')\n\n        def step_I_press_add(self):\n            r\'I press add\'\n\n            raise NotImplementedError(\'I press add\')\n\n        def step_the_result_should_be_number_on_the_screen(self, number):\n            r\'the result should be "([^"]+)" on the screen\'\n\n            raise NotImplementedError(\'the result should be "120" on the screen\')\n\n    ----------------------------------------------------------------------\n    Ran 1 test in 0.013s\n\n    FAILED (failures=1)\n\nNow implement steps with standard `TestCases` that you are familiar:\n\n.. code-block:: python\n\n    # test_acceptance.py\n\n    import unittest\n\n    from morelia import verify\n    \n\n    class CalculatorTestCase(unittest.TestCase):\n    \n        def test_addition(self):\n            """ Addition feature """\n            verify(\'calculator.feature\', self)\n    \n        def step_I_have_powered_calculator_on(self):\n            r\'I have powered calculator on\'\n            self.stack = []\n\n        def step_I_enter_a_number_into_the_calculator(self, number):\n            r\'I enter "(\\d+)" into the calculator\'  # match by regexp\n            self.stack.append(int(number))\n    \n        def step_I_press_add(self):  # matched by method name\n            self.result = sum(self.stack)\n    \n        def step_the_result_should_be_on_the_screen(self, number):\n            r\'the result should be "{number}" on the screen\'  # match by format-like string\n            self.assertEqual(int(number), self.result)\n\n\nAnd run it again:\n\n.. code-block:: console\n\n    $ python -m unittest test_acceptance\n\n    Feature: Addition\n        In order to avoid silly mistakes\n        As a math idiot\n        I want to be told the sum of two numbers\n    Scenario: Add two numbers\n        Given I have powered calculator on                       # pass  0.000s\n        When I enter "50" into the calculator                    # pass  0.000s\n        And I enter "70" into the calculator                     # pass  0.000s\n        And I press add                                          # pass  0.001s\n        Then the result should be "120" on the screen            # pass  0.001s\n    .\n    ----------------------------------------------------------------------\n    Ran 1 test in 0.028s\n\n    OK\n\nNote that Morelia does not waste anyone\'s time inventing a new testing back-end\njust to add a layer of literacy over our testage. Steps are miniature `TestCases`.\nYour onsite customer need never know, and your unit tests and customer tests\ncan share their support methods. The same one test button can run all TDD and BDD tests.\n\nLook at example directory for a little more enhanced example and read full\ndocumentation for more advanced topics.\n\nDocumentation\n=============\n\nFull documentation is available at http://morelia.readthedocs.org/en/latest/index.html\n\n.. image:: http://zeroplayer.com/images/stuff/sneakySnake.jpg\n.. _the cheeseshop: http://pypi.python.org/pypi/Morelia/\n.. _GitHub: http://github.com/kidosoft/Morelia/\n\nCredits\n---------\n\nThis package was created with Cookiecutter_ and the `kidosoft/cookiecutter-pypackage`_ project template.\n\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _`kidosoft/cookiecutter-pypackage`: https://github.com/kidosoft/cookiecutter-pypackage\n',
    'author': 'dryobates',
    'author_email': 'jakub.stolarski@gmail.com',
    'url': 'https://github.com/kidosoft/morelia',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)



from morelia.visitors import VisitorObserver


class Breadcrumbs(VisitorObserver):
    def __init__(self):
        self.__feature = None
        self.__scenario = None
        self.step = None

    def feature_started(self, node):
        self.feature = node

    def scenario_started(self, node):
        self.scenario = node

    def step_started(self, node):
        self.step = node

    @property
    def feature(self):
        return self.__feature

    @feature.setter
    def feature(self, value):
        self.__feature = value
        self.scenario = None

    @property
    def scenario(self):
        return self.__scenario

    @scenario.setter
    def scenario(self, value):
        self.__scenario = value
        self.step = None

    def __str__(self):
        feature = self.feature.interpolated_source() if self.feature else ""
        scenario = self.scenario.interpolated_source() if self.scenario else ""
        step = self.step.interpolated_source() if self.step else ""
        return "\n{}{}{}".format(feature, scenario, step)



import inspect
from abc import ABC
from typing import Iterable, List

from morelia.exceptions import MissingStepError
from morelia.grammar import Feature, Node, Scenario, Step, Visitor


class VisitorObserver(ABC):
    def feature_started(self, node: Feature) -> None:
        self.node_started(node)

    def feature_finished(self, node: Feature) -> None:
        self.node_finished(node)

    def scenario_started(self, node: Scenario) -> None:
        self.node_started(node)

    def scenario_finished(self, node: Scenario) -> None:
        self.node_finished(node)

    def step_started(self, node: Step) -> None:
        self.node_started(node)

    def step_finished(self, node: Step) -> None:
        self.node_finished(node)

    def step_failed(self, node: Step) -> None:
        pass

    def step_errored(self, node: Step) -> None:
        pass

    def node_started(self, node: Node) -> None:
        pass

    def node_finished(self, node: Node) -> None:
        pass


class ObservableVisitor:
    def __init__(self):
        self.__observers = []  # type: List[VisitorObserver]

    def register(self, observer: VisitorObserver) -> None:
        self.__observers.append(observer)

    def feature_started(self, node: Feature) -> None:
        for observer in self.__observers:
            observer.feature_started(node)

    def feature_finished(self, node: Feature) -> None:
        for observer in self.__observers:
            observer.feature_finished(node)

    def scenario_started(self, node: Scenario) -> None:
        for observer in self.__observers:
            observer.scenario_started(node)

    def scenario_finished(self, node: Scenario) -> None:
        for observer in self.__observers:
            observer.scenario_finished(node)

    def step_started(self, node: Step) -> None:
        for observer in self.__observers:
            observer.step_started(node)

    def step_failed(self, node: Step) -> None:
        for observer in self.__observers:
            observer.step_failed(node)

    def step_errored(self, node: Step) -> None:
        for observer in self.__observers:
            observer.step_errored(node)

    def step_finished(self, node: Step) -> None:
        for observer in self.__observers:
            observer.step_finished(node)

    def node_started(self, node: Node) -> None:
        for observer in self.__observers:
            observer.node_started(node)

    def node_finished(self, node: Node) -> None:
        for observer in self.__observers:
            observer.node_finished(node)


class TestVisitor(ObservableVisitor, Visitor):
    """Visits all steps and run step methods."""

    def __init__(self, suite, matcher, scenario_re):
        super().__init__()
        self.__prepare_setup_and_teardown(suite)
        self.__matcher = matcher
        self.__scenario_re = scenario_re

    def __prepare_setup_and_teardown(self, suite):
        self.setUpFeature = getattr(suite, "setUpFeature", self.noop)
        self.tearDownFeature = getattr(suite, "tearDownFeature", self.noop)
        self.setUpScenario = getattr(suite, "setUpScenario", self.noop)
        self.tearDownScenario = getattr(suite, "tearDownScenario", self.noop)
        self.setUpStep = getattr(suite, "setUpStep", self.noop)
        self.tearDownStep = getattr(suite, "tearDownStep", self.noop)

    def noop(self):
        pass

    def visit_feature(self, node: Feature, children: Iterable[Node] = []) -> None:
        self.setUpFeature()
        try:
            self.feature_started(node)
            self.__visit_children(children)
        finally:
            self.tearDownFeature()
            self.feature_finished(node)

    def visit_scenario(self, node: Scenario, children: Iterable[Node] = []) -> None:
        if not self.__scenario_re.match(node.predicate):
            return
        self.setUpScenario()
        try:
            self.scenario_started(node)
            self.__visit_children(children)
        finally:
            self.tearDownScenario()
            self.scenario_finished(node)

    def visit_step(self, node: Step, children: Iterable[Node] = []) -> None:
        self.step_started(node)
        self.setUpStep()
        try:
            self.__execute_step(node)
            self.__visit_children(children)
        except (MissingStepError, AssertionError):
            self.step_failed(node)
            raise
        except (SystemExit, Exception):
            self.step_errored(node)
            raise
        finally:
            self.tearDownStep()
            self.step_finished(node)

    def __execute_step(self, node: Step) -> None:
        __tracebackhide__ = True
        method, args, kwargs = node.find_method(self.__matcher)
        spec = inspect.getfullargspec(method)
        arglist = spec.args + spec.kwonlyargs
        if "_labels" in arglist:
            kwargs["_labels"] = node.get_labels()
        if "_text" in arglist:
            kwargs["_text"] = node.payload
        method(*args, **kwargs)

    def visit(self, node: Node, children: Iterable[Node] = []) -> None:
        try:
            self.node_started(node)
            self.__visit_children(children)
        finally:
            self.node_finished(node)

    visit_background = visit_row = visit_examples = visit_comment = visit

    def __visit_children(self, children: Iterable[Node]) -> None:
        for child in children:
            child.accept(self)
